import tempfile
import os.path as path
from pathlib import Path
import textwrap
import subprocess
import functools
import os
import base64
import re
import sys
import warnings
import shutil
from zipfile import ZipFile, is_zipfile
from pathlib import Path

from lxml import etree
import numpy as np
import click

import gerbonara as gn

@click.group()
def cli():
    pass

@cli.command()
@click.argument('input_gerbers', type=click.Path(exists=True, path_type=Path))
@click.argument('input_svg', type=click.Path(exists=True, dir_okay=False, file_okay=True, allow_dash=True, path_type=Path))
@click.argument('output_gerbers', type=click.Path(allow_dash=True, path_type=Path))
@click.option('--dilate', default=0.1, type=float, help='Default dilation for subtraction operations in mm')
@click.option('--zip/--no-zip', 'is_zip', default=None, help='zip output files. Default: zip if output path ends with ".zip" or when outputting to stdout.')
@click.option('--curve-tolerance', type=float, help='Tolerance for curve flattening in mm')
@click.option('--no-subtract', 'no_subtract', flag_value=True, help='Disable subtraction')
@click.option('--subtract', help='Use user subtraction script from argument')
@click.option('--trace-space', type=float, default=0.1, help='passed through to svg-flatten')
@click.option('--vectorizer', help='passed through to svg-flatten')
@click.option('--vectorizer-map', help='passed through to svg-flatten')
@click.option('--preserve-aspect-ratio', help='PNG/JPG files only: passed through to svg-flatten')
@click.option('--exclude-groups', help='passed through to svg-flatten')
def paste(input_gerbers, input_svg, output_gerbers, is_zip,
        dilate, curve_tolerance, no_subtract, subtract,
        preserve_aspect_ratio,
        trace_space, vectorizer, vectorizer_map, exclude_groups):
    """ Render vector data and raster images from SVG file into gerbers. """

    subtract_map = parse_subtract_script('' if no_subtract else subtract, dilate)

    stack = gn.LayerStack.open(input_gerbers, lazy=True)
    (bb_min_x, bb_min_y), (bb_max_x, bb_max_y) = bounds = stack.board_bounds()

    output_is_zip = output_gerbers.name.lower().endswith('.zip') if is_zip is None else is_zip

    # Create output dir if it does not exist yet. Do this now so we fail early
    if not output_is_zip:
        output_gerbers.mkdir(exist_ok=True)

    @functools.lru_cache()
    def do_dilate(layer, amount):
        return dilate_gerber(layer, bounds, amount, curve_tolerance)
    
    for (side, use), layer in stack.graphic_layers.items():
        overlay_grb = svg_to_gerber(input_svg,
                trace_space=trace_space, vectorizer=vectorizer, vectorizer_map=vectorizer_map,
                exclude_groups=exclude_groups, curve_tolerance=curve_tolerance,
                preserve_aspect_ratio=preserve_aspect_ratio,
                outline_mode=(use == 'outline'),
                only_groups=f'g-{side}-{use}')

        if not overlay_grb:
            print(f'Overlay {side} {use} layer is empty. Skipping.')
            continue

        # only open lazily loaded layer if we need it. Replace lazy wrapper in stack with loaded layer.
        stack.graphic_layers[(side, use)] = layer = layer.instance

        # move overlay from svg origin to gerber origin
        overlay_grb.offset(bb_min_x, bb_min_y)

        # dilated subtract layers on top of overlay
        if side in ('top', 'bottom'): # do not process subtraction scripts for inner layers
            dilations = subtract_map.get(use, [])
            for d_layer, amount in dilations:
                dilated = do_dilate(stack[(side, d_layer)], amount)
                layer.merge(dilated, mode='below', keep_settings=True)

        # overlay on bottom
        layer.merge(overlay_grb, mode='below', keep_settings=True)

    if output_is_zip:
        stack.save_to_zipfile(output_gerbers)
    else:
        stack.save_to_directory(output_gerbers)

@cli.command()
@click.argument('input_gerbers', type=click.Path(exists=True))
@click.argument('output_svg', required=False)
@click.option('-t' ,'--top', help='Render board top side.', is_flag=True)
@click.option('-b' ,'--bottom', help='Render board bottom side.', is_flag=True)
@click.option('-f' ,'--force', help='Overwrite existing output file when autogenerating file name.', is_flag=True)
@click.option('--vector/--raster', help='Embed preview renders into output file as SVG vector graphics instead of rendering them to PNG bitmaps. The resulting preview may slow down your SVG editor.')
@click.option('--raster-dpi', type=float, default=300.0, help='DPI for rastering preview')
def template(input_gerbers, output_svg, top, bottom, force, vector, raster_dpi):
    ''' Generate SVG template for gerbolyze paste from gerber files.

    INPUT may be a gerber file, directory of gerber files or zip file with gerber files
    '''
    source = Path(input_gerbers)
    ttype = 'top' if top else 'bottom'

    if (bool(top) + bool(bottom))  != 1:
        raise click.UsageError('Excactly one of --top or --bottom must be given.')

    if output_svg is None:
        # autogenerate output file name if none is given:
        # /path/to/gerber/dir -> /path/to/gerber/dir.preview-{top|bottom}.svg
        # /path/to/gerbers.zip -> /path/to/gerbers.zip.preview-{top|bottom}.svg
        # /path/to/single/file.grb -> /path/to/single/file.grb.preview-{top|bottom}.svg
    
        output_svg = source.parent / f'{source.name}.template-{ttype}.svg'
        click.echo(f'Writing output to {output_svg}')

        if output_svg.exists() and not force:
            raise UsageError(f'Autogenerated output file already exists. Please remote first, or use --force, or '
                    'explicitly give an output path.')

    else:
        output_svg = Path(output_svg)

    stack = gn.LayerStack.open(source, lazy=True)
    svg = stack.to_pretty_svg(side=('top' if top else 'bottom'), inkscape=True)

    template_layers = [ f'{ttype}-{use}' for use in [ 'copper', 'mask', 'silk' ] ] + ['mechanical outline']
    silk = template_layers[-2]

    if vector:
        output_svg.write_text(create_template_from_svg(svg, template_layers, current_layer=silk))

    else:
        with tempfile.NamedTemporaryFile(suffix='.svg') as temp_svg, \
            tempfile.NamedTemporaryFile(suffix='.png') as temp_png:
            Path(temp_svg.name).write_text(str(svg))
            run_resvg(temp_svg.name, temp_png.name, dpi=f'{raster_dpi:.0f}')
            output_svg.write_text(template_svg_for_png(stack.board_bounds(), Path(temp_png.name).read_bytes(),
                template_layers, current_layer=silk))


class ClickSizeParam(click.ParamType):
    name = 'Size'

    def convert(self, value, param, ctx):
        if isinstance(value, tuple):
            return value

        if not (m := re.match('([0-9]+\.?[0-9]*)(mm|cm|in)?[xX*/,×]([0-9]+\.?[0-9]*)(mm|cm|in)?', value)):
            self.fail('Size must have format [width]x[height][unit]. The unit can be mm, cm or in. The unit is optional and defaults to mm.', param=param, ctx=ctx)

        w, unit1, h, unit2 = m.groups()
        if unit1 and unit2 and unit1 != unit2:
            self.fail('Width and height must use the same unit. Two different units given for width and height: width is in {unit1}, and height is in {unit2}.', param=param, ctx=ctx)

        unit = (unit1 or unit2) or 'mm'
        return float(w), float(h), unit

@cli.command()
@click.argument('output_svg', type=click.Path(dir_okay=False, writable=True, allow_dash=True))
@click.option('--size', type=ClickSizeParam(), default='100x100mm', help='PCB size in [width]x[height][unit] format. Units can be cm, mm or in, when no unit is given, defaults to mm. When no size is given, defaults to 100x100mm.')
@click.option('--force', is_flag=True, help='Overwrite output file without asking if it exists.')
@click.option('-n', '--copper-layers', default=2, type=int, help='Number of copper layers to generate.')
@click.option('--no-default-layers', is_flag=True, help='Do not generate default layers.')
@click.option('-l', '--layer', multiple=True, help='Add given layer to the top of the output layer stack. Can be given multiple times.')
def empty_template(output_svg, size, force, copper_layers, no_default_layers, layer):
    if output_svg == '-':
        out = sys.stdout
    else:
        out = Path(output_svg)
        if out.exists():
            if not force and not click.confirm(f'Output file "{out}" already exists. Do you want to overwrite it?'):
                raise click.ClickException(f'Output file "{out}" already exists, exiting.')
        out = out.open('w')

    layers = layer or []
    current_layer = None
    if not no_default_layers:
        layers += ['top paste', 'top silk', 'top mask']

        if copper_layers > 0:
            current_layer = 'top copper'
            inner = [ 'inner{i} copper' for i in range(max(0, copper_layers-2)) ]
            layers += ['top copper', *inner, 'bottom copper'][:copper_layers]

        layers += ['bottom mask', 'bottom silk', 'bottom paste']
        layers += ['outline', 'plated drill', 'nonplated drill', 'comments']
    if layers and current_layer is None:
        current_layer = layers[0]

    out.write(empty_pcb_template(size, layers, current_layer))
    out.flush()
    if output_svg != '-':
        out.close()


@cli.command()
@click.argument('input_svg', type=click.Path(exists=True, path_type=Path))
@click.argument('output_gerbers', type=click.Path(path_type=Path))
@click.option('-n', '--naming-scheme', default='kicad', type=click.Choice(['kicad', 'altium']), help='Naming scheme for gerber output file names.')
@click.option('--zip/--no-zip', 'is_zip', default=None, help='zip output files. Default: zip if output path ends with ".zip" or when outputting to stdout.')
@click.option('--separate-drill-file/--composite-drill-file', 'separate_drill', help='Use Altium composite Excellon drill file format (default)')
@click.option('--dilate', default=0.1, type=float, help='Default dilation for subtraction operations in mm')
@click.option('--curve-tolerance', type=float, help='Tolerance for curve flattening in mm')
@click.option('--no-subtract', 'no_subtract', flag_value=True, help='Disable subtraction')
@click.option('--subtract', help='Use user subtraction script from argument (see description above)')
@click.option('--trace-space', type=float, default=0.1, help='passed through to svg-flatten')
@click.option('--vectorizer', help='passed through to svg-flatten')
@click.option('--vectorizer-map', help='passed through to svg-flatten')
@click.option('--exclude-groups', help='passed through to svg-flatten')
@click.option('--pattern-complete-tiles-only', is_flag=True, help='passed through to svg-flatten')
@click.option('--use-apertures-for-patterns', is_flag=True, help='passed through to svg-flatten')
def convert(input_svg, output_gerbers, is_zip, dilate, curve_tolerance, no_subtract, subtract, trace_space, vectorizer,
        vectorizer_map, exclude_groups, separate_drill, naming_scheme,
        pattern_complete_tiles_only, use_apertures_for_patterns):
    ''' Convert SVG file directly to gerbers.

    Unlike `gerbolyze paste`, this does not add the SVG's contents to existing gerbers. It allows you to directly create
    PCBs using Inkscape similar to PCBModE.
    '''

    subtract_map = parse_subtract_script('' if no_subtract else subtract, dilate, default_script=DEFAULT_CONVERT_SUB_SCRIPT)
    output_is_zip = output_gerbers.name.lower().endswith('.zip') if is_zip is None else is_zip

    stack = gn.LayerStack({}, [], board_name=input_svg.stem, original_path=input_svg)

    for group_id, label in get_layers_from_svg(input_svg.read_text()):
        if not group_id or not label or 'no export' in label:
            continue

        if label == 'outline':
            side, use = 'mechanical', 'outline'
        elif label == 'comments':
            side, use = 'other', 'comments'
        elif len(label.split()) != 2:
            warnings.warn('Unknown layer {label}')
            continue
        else:
            side, use = label.split()

        grb = svg_to_gerber(input_svg,
                trace_space=trace_space, vectorizer=vectorizer, vectorizer_map=vectorizer_map,
                exclude_groups=exclude_groups, curve_tolerance=curve_tolerance, only_groups=group_id,
                pattern_complete_tiles_only=pattern_complete_tiles_only,
                use_apertures_for_patterns=(use_apertures_for_patterns and use not in ('outline', 'drill')),
                outline_mode=(use == 'outline' or use == 'drill'))
        grb.original_path = Path()

        if use == 'drill':
            if side == 'plated':
                stack.drill_pth = grb.to_excellon(plated=True)
            elif side == 'nonplated':
                stack.drill_npth = grb.to_excellon(plated=False)
            else:
                warnings.warn(f'Invalid drill layer type "{side}". Must be one of "plated" or "nonplated"')

        else:
            stack.graphic_layers[(side, use)] = grb

    bounds = stack.board_bounds()
    @functools.lru_cache()
    def do_dilate(layer, amount):
        return dilate_gerber(layer, bounds, amount, curve_tolerance)

    for (side, use), layer in stack.graphic_layers.items():
        # dilated subtract layers on top of overlay
        if side in ('top', 'bottom'): # do not process subtraction scripts for inner layers
            dilations = subtract_map.get(use, [])
            for d_layer, amount in dilations:
                d_layer =  stack.graphic_layers[(side, d_layer)]
                dilated = do_dilate(d_layer, amount)
                layer.merge(dilated, mode='above', keep_settings=True)

    if not separate_drill:
        print('Merging drill layers...')
        stack.merge_drill_layers()

    naming_scheme = getattr(gn.layers.NamingScheme, naming_scheme)
    if output_is_zip:
        stack.save_to_zipfile(output_gerbers, naming_scheme=naming_scheme)
    else:
        stack.save_to_directory(output_gerbers, naming_scheme=naming_scheme)



# Subtraction script handling
#============================

DEFAULT_SUB_SCRIPT = '''
out.silk -= in.mask
out.silk -= in.silk+0.5
out.mask -= in.mask+0.5
out.copper -= in.copper+0.5
'''

DEFAULT_CONVERT_SUB_SCRIPT = '''
out.silk -= in.mask
'''

def parse_subtract_script(script, default_dilation=0.1, default_script=DEFAULT_SUB_SCRIPT):
    if script is None:
        script = default_script

    subtract_script = {}
    lines = script.replace(';', '\n').splitlines()
    for line in lines:
        line = line.strip()
        if not line or line.startswith('#'):
            continue

        line = line.lower()
        line = re.sub('\s', '', line)

        # out.copper -= in.copper+0.1
        varname = r'([a-z]+\.[a-z]+)'
        floatnum = r'([+-][.0-9]+)'
        match = re.fullmatch(fr'{varname}-={varname}{floatnum}?', line)
        if not match:
            raise ValueError(f'Cannot parse line: {line}')
        
        out_var, in_var, dilation = match.groups()
        if not out_var.startswith('out.') or not in_var.startswith('in.'):
            raise ValueError('All left-hand side values must be outputs, right-hand side values must be inputs.')

        _out, _, out_layer = out_var.partition('.')
        _in, _, in_layer = in_var.partition('.')

        dilation = float(dilation) if dilation else default_dilation

        subtract_script[out_layer] = subtract_script.get(out_layer, []) + [(in_layer, dilation)]
    return subtract_script

# Parameter parsing foo
#======================

def parse_bbox(bbox):
    if not bbox:
        return None
    elems = [ int(elem) for elem in re.split('[,/ ]', bbox) ]
    if len(elems) not in (2, 4):
        raise click.BadParameter(
                '--bbox must be either two floating-point values like: w,h or four like: x,y,w,h')

    elems = [ float(e) for e in elems ]

    if len(elems) == 2:
        bounds = [0, 0, *elems]
    else:
        bounds = elems
    
    # now transform bounds to the format pcb-tools uses. Instead of (x, y, w, h) or even (x1, y1, x2, y2), that
    # is ((x1, x2), (y1, y2)

    x, y, w, h = bounds
    return ((x, x+w), (y, y+h))

# Utility foo
# ===========

def run_resvg(input_file, output_file, **resvg_args):

    args = []
    for key, value in resvg_args.items():
        if value is not None:
            if value is False:
                continue

            args.append(f'--{key.replace("_", "-")}')

            if value is not True:
                args.append(value)

    args += [input_file, output_file]

    # By default, try a number of options:
    candidates = [
        # somewhere in $PATH
        'resvg',
        'wasi-resvg',
        # in user-local cargo installation
        Path.home() / '.cargo' / 'bin' / 'resvg',
        # wasi-resvg in user-local pip installation
        Path.home() / '.local' / 'bin' / 'wasi-resvg',
        # next to our current python interpreter (e.g. in virtualenv)
        str(Path(sys.executable).parent / 'wasi-resvg')
        ]

    # if RESVG envvar is set, try that first.
    if 'RESVG' in os.environ:
        candidates = [os.environ['RESVG'], *candidates]

    for candidate in candidates:
        try:
            res = subprocess.run([candidate, *args], check=True)
            print('used resvg:', candidate)
            break
        except FileNotFoundError:
            continue
    else:
        raise SystemError('resvg executable not found')



def calculate_apertureless_bounding_box(cam):
    ''' pcb-tools'es default bounding box function returns the bounding box of the primitives including apertures (i.e.
    line widths). For determining a board's size from the outline layer, we want the bounding box disregarding
    apertures.
    '''

    min_x = min_y = 1000000
    max_x = max_y = -1000000

    for prim in cam.primitives:
        bounds = prim.bounding_box_no_aperture
        min_x = min(bounds[0][0], min_x)
        max_x = max(bounds[0][1], max_x)

        min_y = min(bounds[1][0], min_y)
        max_y = max(bounds[1][1], max_y)

    return ((min_x, max_x), (min_y, max_y))

# SVG export
#===========

def template_layer(name):
    return f'<g id="g-{name.lower().replace(" ", "-")}" inkscape:label="{name}" inkscape:groupmode="layer"></g>'

def template_svg_for_png(bounds, png_data, extra_layers, current_layer):
    (x1, y1), (x2, y2) = bounds
    w_mm, h_mm = (x2 - x1), (y2 - y1)

    extra_layers = "\n  ".join(template_layer(name) for name in extra_layers)

    # we set up the viewport such that document dimensions = document units = mm
    template = f'''<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <svg version="1.1"
           xmlns="http://www.w3.org/2000/svg"
           xmlns:xlink="http://www.w3.org/1999/xlink"
           xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
           xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
           width="{w_mm}mm" height="{h_mm}mm" viewBox="0 0 {w_mm} {h_mm}" >
          <defs/>
          <sodipodi:namedview inkscape:current-layer="g-{current_layer.lower().replace(" ", "-")}" />
          <g inkscape:label="Preview" inkscape:groupmode="layer" id="g-preview" sodipodi:insensitive="true" style="opacity:0.5">
            <image x="0" y="0" width="{w_mm}" height="{h_mm}"
               xlink:href="data:image/jpeg;base64,{base64.b64encode(png_data).decode()}" />
          </g>
          {extra_layers}
        </svg>
        '''
    return textwrap.dedent(template)

def empty_pcb_template(size, extra_layers, current_layer):
    w, h, unit = size

    extra_layers = "\n  ".join(template_layer(name) for name in extra_layers)
    current_layer = f'<sodipodi:namedview inkscape:current-layer="g-{current_layer.lower().replace(" ", "-")}" />' if current_layer else ''

    # we set up the viewport such that document dimensions = document units = [unit]
    template = f'''<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <svg version="1.1"
           xmlns="http://www.w3.org/2000/svg"
           xmlns:xlink="http://www.w3.org/1999/xlink"
           xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
           xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
           width="{w}{unit}" height="{h}{unit}" viewBox="0 0 {w} {h}" >
          <defs/>
          {current_layer}
          {extra_layers}
        </svg>
        '''
    return textwrap.dedent(template)


MM_PER_INCH = 25.4

def create_template_from_svg(svg, extra_layers, current_layer):
    view, *layers = svg.children
    view.attrs['inkscape__current_layer'] = f'g-{current_layer.lower().replace(" ", "-")}'

    extra_layers = [ template_layer(name) for name in extra_layers ]
    svg.children = [ view, *extra_layers, gn.utils.Tag('g', layers, inkscape__label='Preview', sodipodi__insensitive='true',
        inkscape__groupmode='layer', style='opacity:0.5') ]

    return str(svg)

# SVG/gerber import
#==================

def dilate_gerber(layer, bounds, dilation, curve_tolerance):
    with tempfile.NamedTemporaryFile(suffix='.svg') as temp_svg:
        Path(temp_svg.name).write_text(str(layer.instance.to_svg(force_bounds=bounds, fg='white')))

        # dilate & render back to gerber
        # NOTE: Maybe reconsider or nicely document dilation semantics ; It is weird that negative dilations affect
        # clear color and positive affects dark colors
        out = svg_to_gerber(temp_svg.name, dilate=-dilation, curve_tolerance=curve_tolerance)
        return out

def svg_to_gerber(infile, outline_mode=False, **kwargs):
    infile = Path(infile)

    args = [ '--format', ('gerber-outline' if outline_mode else 'gerber'),
            '--precision', '6', # intermediate file, use higher than necessary precision
            ]
    
    for k, v in kwargs.items():
        if v:
            args.append('--' + k.replace('_', '-'))
            if not isinstance(v, bool):
                args.append(str(v))

    with tempfile.NamedTemporaryFile(suffix='.gbr') as temp_gbr:
        args += [str(infile), str(temp_gbr.name)]
        print(' '.join(args))

        if 'SVG_FLATTEN' in os.environ:
            subprocess.run([os.environ['SVG_FLATTEN'], *args], check=True)
            print('used svg-flatten at $SVG_FLATTEN')

        else:
            # By default, try four options:
            for candidate in [
                    # somewhere in $PATH
                    'svg-flatten',
                    None, # direct WASI import
                    'wasi-svg-flatten',

                    # in user-local pip installation
                    Path.home() / '.local' / 'bin' / 'svg-flatten',
                    Path.home() / '.local' / 'bin' / 'wasi-svg-flatten',

                    # next to our current python interpreter (e.g. in virtualenv)
                    str(Path(sys.executable).parent / 'svg-flatten'),
                    str(Path(sys.executable).parent / 'wasi-svg-flatten'),

                    # next to this python source file in the development repo
                    str(Path(__file__).parent.parent / 'svg-flatten' / 'build' / 'svg-flatten') ]:

                try:
                    if candidate is None:
                        import svg_flatten_wasi
                        svg_flatten_wasi.run_svg_flatten.callback(args[-2], args[-1], args[:-2], no_usvg=False)
                        print('used svg_flatten_wasi python package') 

                    else:
                        subprocess.run([candidate, *args], check=True)
                        print('used svg-flatten at', candidate)

                    break
                except (FileNotFoundError, ModuleNotFoundError):
                    continue

            else:
                raise SystemError('svg-flatten executable not found')

        return gn.rs274x.GerberFile.open(temp_gbr.name)

def get_layers_from_svg(svg_data):
    svg = etree.fromstring(svg_data.encode('utf-8'))
    SVG_NS = '{http://www.w3.org/2000/svg}'
    INKSCAPE_NS = '{http://www.inkscape.org/namespaces/inkscape}'

    # find groups
    for group in svg.findall(SVG_NS+'g'):
        yield group.get('id'), group.get(INKSCAPE_NS+'label')
   

if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
#
# Copyright 2022 Jan Götte <code@jaseg.de>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import subprocess
import tempfile
from pathlib import Path

import pytest


REFERENCE_GERBERS = ['test_gerber_8seg.zip']
REFERENCE_SVGS = ['svg_feature_test.svg']

reference_path = lambda reference: Path(__file__).parent / 'resources' / str(reference)


def run_command(*args):
    try:
        proc = subprocess.run(args, check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        print(e.stdout.decode())
        print(e.stderr.decode(), file=sys.stderr)
        raise

def test_template_round_trip():
    with tempfile.NamedTemporaryFile(suffix='.svg') as out_svg,\
            tempfile.TemporaryDirectory() as out_dir:
        run_command('python3', '-m', 'gerbolyze', 'empty-template', '--force', out_svg.name)
        run_command('python3', '-m', 'gerbolyze', 'convert', out_svg.name, out_dir)

def test_zip_write():
    with tempfile.NamedTemporaryFile(suffix='.svg') as out_svg,\
            tempfile.NamedTemporaryFile(suffix='.zip') as out_zip:
        run_command('python3', '-m', 'gerbolyze', 'empty-template', '--force', out_svg.name)
        run_command('python3', '-m', 'gerbolyze', 'convert', out_svg.name, out_zip.name)

@pytest.mark.parametrize('reference', REFERENCE_SVGS)
def test_complex_conversion(reference):
    infile = reference_path(reference)
    with tempfile.NamedTemporaryFile(suffix='.zip') as out_zip:
        run_command('python3', '-m', 'gerbolyze', 'convert', infile, out_zip.name)
        run_command('python3', '-m', 'gerbolyze', 'convert', '--pattern-complete-tiles-only', '--use-apertures-for-patterns', infile, out_zip.name)

@pytest.mark.parametrize('reference', REFERENCE_GERBERS)
def test_template(reference):
    with tempfile.NamedTemporaryFile(suffix='.zip') as out_svg:
        infile = reference_path(reference)
        run_command('python3', '-m', 'gerbolyze', 'template', '--top', '--force', infile, out_svg.name)
        run_command('python3', '-m', 'gerbolyze', 'template', '--bottom', '--force', '--vector', infile, out_svg.name)




#!/usr/bin/env python3

import re
import textwrap
import ast
import uuid

svg_str = lambda content: content if isinstance(content, str) else '\n'.join(str(c) for c in content)

class Pattern:
    def __init__(self, w, h=None):
        self.vb_w = self.w = w
        self.vb_h = self.h = h or w

    def svg_def(self, svg_id, off_x, off_y):
        return textwrap.dedent(f'''
            <pattern id="{svg_id}" x="{off_x}" y="{off_y}" viewBox="0,0,{self.vb_w},{self.vb_h}" width="{self.w}" height="{self.h}" patternUnits="userSpaceOnUse">
                {svg_str(self.content)}
            </pattern>''')
    
def make_rect(svg_id, x, y, w, h, clip=''):
    #import random
    #c = random.randint(0, 2**24)
    #return f'<rect x="{x}" y="{y}" width="{w}" height="{h}" fill="#{c:06x}"/>'
    return f'<rect x="{x}" y="{y}" width="{w}" height="{h}" {clip} fill="url(#{svg_id})"/>'

class CirclePattern(Pattern):
    def __init__(self, d, w, h=None):
        super().__init__(w, h)
        self.d = d

    @property
    def content(self):
        return f'<circle cx="{self.w/2}" cy="{self.h/2}" r="{self.d/2}"/>'

class RectPattern(Pattern):
    def __init__(self, rw, rh, w, h):
        super().__init__(w, h)
        self.rw, self.rh = rw, rh

    @property
    def content(self):
        x = (self.w - self.rw) / 2
        y = (self.h - self.rh) / 2
        return f'<rect x="{x}" y="{y}" width="{self.rw}" height="{self.rh}"/>'

class ManhattanPattern(Pattern):
    def __init__(self, pitch=2.54*4, gap=0.2):
        super().__init__(pitch)
        self.vb_w, self.vb_h = 1, 1
        self.gap = gap

    @property
    def content(self):
        return textwrap.dedent('''
                <rect x="0"   y="0"   width="0.5" height="0.5" style="fill: black; stroke: white; stroke-width: 0.01mm"/>
                <rect x="0"   y="0.5"   width="0.5" height="0.5" style="fill: black; stroke: white; stroke-width: 0.01mm"/>
                <rect x="0.5"   y="0"   width="0.5" height="0.5" style="fill: black; stroke: white; stroke-width: 0.01mm"/>
                <rect x="0.5"   y="0.5"   width="0.5" height="0.5" style="fill: black; stroke: white; stroke-width: 0.01mm"/>
                <rect x="0.3" y="0.3" width="0.4" height="0.4" style="fill: black; stroke: white; stroke-width: 0.01mm" transform="rotate(45 0.5 0.5)"/>
                '''.strip())

make_layer = lambda layer_name, content: \
  f'<g id="g-{layer_name.replace(" ", "-")}" inkscape:label="{layer_name}" inkscape:groupmode="layer">{svg_str(content)}</g>'

svg_template = textwrap.dedent('''
    <?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <svg version="1.1" width="{w}mm" height="{h}mm" viewBox="0 0 {w} {h}" id="svg18" sodipodi:docname="proto.svg"
       inkscape:version="1.2 (dc2aedaf03, 2022-05-15)"
       xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
       xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
       xmlns="http://www.w3.org/2000/svg"
       xmlns:svg="http://www.w3.org/2000/svg">
      <defs id="defs2">
        {defs}
      </defs>
      <sodipodi:namedview inkscape:current-layer="g-top-copper" id="namedview4" pagecolor="#ffffff" bordercolor="#666666"
         borderopacity="1.0" inkscape:showpageshadow="2" inkscape:pageopacity="0.0" inkscape:pagecheckerboard="0"
         inkscape:deskcolor="#d1d1d1" inkscape:document-units="mm" showgrid="false" inkscape:zoom="2.8291492"
         inkscape:cx="157.29111" inkscape:cy="80.943063" inkscape:window-width="1920" inkscape:window-height="1011"
         inkscape:window-x="0" inkscape:window-y="0" inkscape:window-maximized="1" />
      {layers}
    </svg>
''').strip()

class PatternProtoArea:
    def __init__(self, pitch_x, pitch_y=None, border=None):
        self.pitch_x = pitch_x
        self.pitch_y = pitch_y or pitch_x

        if border is None:
            self.border = (0, 0, 0, 0)
        elif hasattr(border, '__iter__'):
            if len(border == 4):
                self.border = border
            else:
                raise TypeError('border must be None, int, or a 4-tuple of floats (top, right, bottom, left)')
        else:
            self.border = (border, border, border, border)

    @property
    def pitch(self):
        if self.pitch_x != self.pitch_y:
            raise ValueError('Pattern has different X and Y pitches')
        return self.pitch_x

    def fit_size(self, w, h):
        x, y, w, h = self.fit_rect(0, 0, w, h, False)
        t, r, b, l = self.border
        return (w+l+r), (h+t+b)

    def fit_rect(self, x, y, w, h, center=True):
        t, r, b, l = self.border
        x, y, w, h = (x+l), (y+t), (w-l-r), (h-t-b)

        w_mod, h_mod = round((w + 5e-7) % self.pitch_x, 6), round((h + 5e-7) % self.pitch_y, 6)
        w_fit, h_fit = round(w - w_mod, 6), round(h - h_mod, 6)

        if center:
            x = x + (w-w_fit)/2
            y = y + (h-h_fit)/2
            return x, y, w_fit, h_fit

        else:
            return x, y, w_fit, h_fit

    def generate(self, x, y, w, h, center=True, clip='', tight_layout=False):
        yield {}

    def symmetric_sides(self):
        return False

    def used_patterns(self):
        yield self


class EmptyProtoArea:
    def __init__(self, copper=False, border=None):
        self.copper = copper

        if border is None:
            self.border = (0, 0, 0, 0)
        elif hasattr(border, '__iter__'):
            if len(border == 4):
                self.border = border
            else:
                raise TypeError('border must be None, int, or a 4-tuple of floats (top, right, bottom, left)')
        else:
            self.border = (border, border, border, border)

    def fit_size(self, w, h):
        return w, h

    def generate(self, x, y, w, h, center=True, clip='', tight_layout=False):
        if self.copper:
            t, r, b, l = self.border
            x, y, w, h = x+l, y+t, w-l-r, h-t-b
            yield { 'top copper': f'<rect x="{x}" y="{y}" width="{w}" height="{h}" {clip} fill="black"/>' }
        else:
            yield {}

    def used_patterns(self):
        yield self


class THTProtoArea(PatternProtoArea):
    def __init__(self, pad_size=2.0, drill=1.0, pitch=2.54, sides='both', plated=True, border=None, pad_shape='circle'):
        super().__init__(pitch, border=border)
        self.pad_size = pad_size
        self.pad_shape = pad_shape.lower().rstrip('s')
        self.drill = drill
        self.drill_pattern = CirclePattern(self.drill, self.pitch)
        if self.pad_shape == 'circle':
            self.pad_pattern = CirclePattern(self.pad_size, self.pitch)
        elif self.pad_shape == 'square':
            self.pad_pattern = RectPattern(self.pad_size, self.pad_size, self.pitch, self.pitch)
        self.patterns = [self.drill_pattern, self.pad_pattern]
        self.plated = plated
        self.sides = sides
    
    def generate(self, x, y, w, h, center=True, clip='', tight_layout=False):
        x, y, w, h = self.fit_rect(x, y, w, h, center)
        drill = 'plated drill' if self.plated else 'nonplated drill'

        pad_id = str(uuid.uuid4())
        drill_id = str(uuid.uuid4())

        d = { drill: make_rect(drill_id, x, y, w, h, clip),
             'defs': [
                 self.pad_pattern.svg_def(pad_id, x, y),
                 self.drill_pattern.svg_def(drill_id, x, y)]}

        if self.sides in ('top', 'both'):
            d['top copper'] = make_rect(pad_id, x, y, w, h, clip)
            d['top mask'] = make_rect(pad_id, x, y, w, h, clip)
        if self.sides in ('bottom', 'both'):
            d['bottom copper'] = make_rect(pad_id, x, y, w, h, clip)
            d['bottom mask'] = make_rect(pad_id, x, y, w, h, clip)

        yield d

    def __repr__(self):
        return f'THTPads(size={self.pad_size}, h={self.drill}, p={self.pitch}, sides={self.sides}, plated={self.plated}, pad_shape="{self.pad_shape}")'

    def symmetric_sides(self):
        return True


class SMDProtoAreaRectangles(PatternProtoArea):
    def __init__(self, pitch_x, pitch_y, w=None, h=None, border=None):
        super().__init__(pitch_x, pitch_y, border=border)
        w = w or pitch_x - 0.15
        h = h or pitch_y - 0.15
        self.w, self.h = w, h
        self.pad_pattern = RectPattern(w, h, pitch_x, pitch_y)
        self.patterns = [self.pad_pattern]

    def generate(self, x, y, w, h, center=True, clip='', tight_layout=False):
        x, y, w, h = self.fit_rect(x, y, w, h, center)
        pad_id = str(uuid.uuid4())
        yield {'defs': [self.pad_pattern.svg_def(pad_id, x, y)],
            'top copper': make_rect(pad_id, x, y, w, h, clip),
            'top mask': make_rect(pad_id, x, y, w, h, clip)}

    def symmetric_sides(self):
        return False

class ManhattanProtoArea(PatternProtoArea):
    def __init__(self, pitch=2.54*4, gap=0.25, border=None):
        super().__init__(pitch, pitch, border=border)
        self.gap = gap
        self.pad_pattern = ManhattanPattern(pitch, gap)
        self.patterns = [self.pad_pattern]

    def generate(self, x, y, w, h, center=True, clip='', tight_layout=False):
        x, y, w, h = self.fit_rect(x, y, w, h, center)
        pad_id = str(uuid.uuid4())
        yield {'defs': [self.pad_pattern.svg_def(pad_id, x, y)],
            'top copper': make_rect(pad_id, x, y, w, h, clip),
            'top mask': make_rect(pad_id, x, y, w, h, clip)}

    def symmetric_sides(self):
        return False

LAYERS = [
        'top paste',
        'top silk',
        'top mask',
        'top copper',
        'bottom copper',
        'bottom mask',
        'bottom silk',
        'bottom paste',
        'outline',
        'nonplated drill',
        'plated drill'
        ]

class ProtoBoard:
    def __init__(self, defs, expr, mounting_holes=None, border=None, center=True, tight_layout=False):
        self.defs = eval_defs(defs)
        self.layout = parse_layout(expr, self.defs)
        self.mounting_holes = mounting_holes
        self.center = center
        self.tight_layout = tight_layout

        if border is None:
            self.border = (0, 0, 0, 0)
        elif hasattr(border, '__iter__'):
            if len(border == 4):
                self.border = border
            else:
                raise TypeError('border must be None, int, or a 4-tuple of floats (top, right, bottom, left)')
        else:
            self.border = (border, border, border, border)

    @property
    def symmetric_sides(self):
        return self.layout.symmetric_sides()

    @property
    def used_patterns(self):
        return set(self.layout.used_patterns())

    def generate(self, w, h):
        out = {l: [] for l in LAYERS}
        svg_defs = []
        clip = ''

        if self.mounting_holes:
            d, o, *k = self.mounting_holes # diameter, offset from edge, keepout to proto area
            k = k[0] if k else o
            q = o + k
            if 2*q < w:
                if 2*q < h:
                    clip_d = f'M 0 {q} L {q} {q} L {q} 0 L {w-q} 0 L {w-q} {q} L {w} {q} L {w} {h-q} L {w-q} {h-q} L {w-q} {h} L {q} {h} L {q} {h-q} L 0 {h-q} Z'
                else:
                    clip_d = f'M {q} 0 L {w-q} 0 L {w-q} {h} L 0 {h} Z'
            else:
                if 2*q < h:
                    clip_d = f'M 0 {q} L 0 {h-q} L {w} {h-q} L {w} {q} Z'
                else:
                    raise ValueError(f'Hole keepout areas are so large that no board area is left. Available size is {w}x{h} mm, keepout areas are {q}x{q} mm in all four corners.')

            svg_defs.append(f'<clipPath id="hole-clip"><path d="{clip_d}"/></clipPath>')
            clip = 'clip-path="url(#hole-clip)"'

            out['nonplated drill'].append([
                f'<circle cx="{o}" cy="{o}" r="{d/2}"/>',
                f'<circle cx="{w-o}" cy="{o}" r="{d/2}"/>',
                f'<circle cx="{w-o}" cy="{h-o}" r="{d/2}"/>',
                f'<circle cx="{o}" cy="{h-o}" r="{d/2}"/>' ])

        t, r, b, l = self.border
        for layer_dict in self.layout.generate(l, t, w-l-r, h-t-b, self.center, clip, self.tight_layout):
            for l in LAYERS:
                if l in layer_dict:
                    out[l].append(layer_dict[l])
            svg_defs += layer_dict.get('defs', [])

        out['outline'] = f'<rect x="0" y="0" width="{w}" height="{h}" fill="none" stroke="black" stroke-width="0.1mm"/>'

        layers = [ make_layer(l, out[l]) for l in LAYERS ]
        return svg_template.format(w=w, h=h, defs='\n'.join(svg_defs), layers='\n'.join(layers)) 


def convert_to_mm(value, unit):
    unitl  = unit.lower()
    if unitl == 'mm':
        return value
    elif unitl == 'cm':
        return value*10
    elif unitl == 'in':
        return value*25.4
    elif unitl == 'mil':
        return value/1000*25.4
    else:
        raise ValueError(f'Invalid unit {unit}, allowed units are mm, cm, in, and mil.')

value_re = re.compile('([0-9]*\.?[0-9]+)(cm|mm|in|mil|%)')
def eval_value(value, total_length=None):
    if not isinstance(value, str):
        return None

    m = value_re.match(value.lower())
    number, unit = m.groups()
    if unit == '%':
        if total_length is None:
            raise ValueError('Percentages are not allowed for this value')
        return total_length * float(number) / 100
    return convert_to_mm(float(number), unit)

class PropLayout:
    def __init__(self, content, direction, proportions):
        self.content = content
        self.direction = direction
        self.proportions = proportions
        if len(content) != len(proportions):
            raise ValueError('proportions and content must have same length')

    def generate(self, x, y, w, h, center=True, clip='', tight_layout=False):
        for (c_x, c_y, c_w, c_h), child in self.layout_2d(x, y, w, h, tight_layout):
            yield from child.generate(c_x, c_y, c_w, c_h, center, clip, tight_layout)

    def fit_size(self, w, h):
        widths = []
        heights = []
        for (_x, _y, w, h), child in self.layout_2d(0, 0, w, h, True):
            if not isinstance(child, EmptyProtoArea):
                widths.append(w)
                heights.append(h)
        if self.direction == 'h':
            return sum(widths), max(heights)
        else:
            return max(widths), sum(heights)

    def layout_2d(self, x, y, w, h, tight_layout=False):
        actual_l = 0
        target_l = 0
        for l, child in zip(self.layout(w if self.direction == 'h' else h), self.content):
            this_x, this_y = x, y
            this_w, this_h = w, h
            target_l += l

            if self.direction == 'h':
                this_w = target_l - actual_l
            else:
                this_h = target_l - actual_l

            if tight_layout:
                this_w, this_h = child.fit_size(this_w, this_h)

            if self.direction == 'h':
                x += this_w
                actual_l += this_w
                this_h = h
            else:
                y += this_h
                actual_l += this_h
                this_w = w

            yield (this_x, this_y, this_w, this_h), child

    def layout(self, length):
        out = [ eval_value(value, length) for value in self.proportions ]
        total_length = sum(value for value in out if value is not None)
        if length - total_length < -1e-6:
            raise ValueError(f'Proportions sum to {total_length} mm, which is greater than the available space of {length} mm.')

        leftover = length - total_length
        sum_props = sum( (value or 1.0) for value in self.proportions if not isinstance(value, str) )
        return [ (leftover * (value or 1.0) / sum_props if not isinstance(value, str) else calculated)
                for value, calculated in zip(self.proportions, out) ]

    def __str__(self):
        children = ', '.join( f'{elem}:{width}' for elem, width in zip(self.content, self.proportions))
        return f'PropLayout[{self.direction.upper()}]({children})'

    def symmetric_sides(self):
        return all(child.symmetric_sides() for child in self.content)

    def used_patterns(self):
        for child in self.content:
            yield from child.used_patterns()


class TwoSideLayout:
    def __init__(self, top, bottom):
        self.top, self.bottom = top, bottom

    def flip(self, defs):
        out = dict(defs)
        for layer in ('copper', 'mask', 'silk', 'paste'):
            top, bottom = f'top {layer}', f'bottom {layer}'
            tval, bval = defs.get(top), defs.get(bottom)

            if tval:
                defs[bottom] = tval
            elif bottom in defs:
                del defs[bottom]

            if bval:
                defs[top] = bval
            elif top in defs:
                del defs[top]

        return defs

    def fit_size(self, w, h):
        top, bottom = self.top, self.bottom
        w1, h1 = top.fit_size(w, h)
        w2, h2 = bottom.fit_size(w, h)
        if isinstance(top, EmptyProtoArea):
            if isinstance(bottom, EmptyProtoArea):
                return w1, h1
            return w2, h2
        if isinstance(bottom, EmptyProtoArea):
            return w1, h1
        return max(w1, w2), max(h1, h2)

    def generate(self, x, y, w, h, center=True, clip='', tight_layout=False):
        yield from self.top.generate(x, y, w, h, center, clip, tight_layout)
        yield from map(self.flip, self.bottom.generate(x, y, w, h, center, clip, tight_layout))

    def symmetric_sides(self):
        return self.top == self.bottom

    def used_patterns(self):
        yield from self.top.used_patterns()
        yield from self.bottom.used_patterns()


def _map_expression(node, defs):
    if isinstance(node, ast.Name):
        return defs[node.id]

    elif isinstance(node, ast.Constant):
        return node.value


    elif isinstance(node, ast.BinOp) and isinstance(node.op, (ast.BitOr, ast.BitAnd, ast.Add)):
        left_prop = right_prop = None

        left, right = node.left, node.right

        if isinstance(left, ast.BinOp) and isinstance(left.op, ast.MatMult):
            left_prop = _map_expression(left.right, defs)
            left = left.left

        if isinstance(right, ast.BinOp) and isinstance(right.op, ast.MatMult):
            right_prop = _map_expression(right.right, defs)
            right = right.left

        left, right = _map_expression(left, defs), _map_expression(right, defs)

        direction = 'h' if isinstance(node.op, ast.BitOr) else 'v'
        if isinstance(left, PropLayout) and left.direction == direction and left_prop is None:
            left.content.append(right)
            left.proportions.append(right_prop)
            return left

        elif isinstance(right, PropLayout) and right.direction == direction and right_prop is None:
            right.content.insert(0, left)
            right.proportions.insert(0, left_prop)
            return right

        elif isinstance(node.op, ast.Add):
            if left_prop or right_prop:
                raise SyntaxError(f'Proportions ("@") not supported for two-side layout ("+")')

            return TwoSideLayout(left, right)

        else:
            return PropLayout([left, right], direction, [left_prop, right_prop])
        
    elif isinstance(node, ast.BinOp) and isinstance(node.op, ast.MatMult):
        raise SyntaxError(f'Unexpected width specification "{ast.unparse(node.right)}"')

    else:
        raise SyntaxError(f'Invalid layout expression "{ast.unparse(node)}"')

def parse_layout(expr, defs):
    ''' Example layout:

        ( tht @ 2in | smd ) @ 50% / tht
    '''

    expr = re.sub(r'\s', '', expr)
    expr = re.sub(r'([0-9]*\.?[0-9]+)([Mm][Mm]|[Cc][Mm]|[Ii][Nn]|[Mm][Ii][Ll]|%)', r'"\1\2"', expr)
    expr = expr.replace('/', '&')
    try:
        expr = ast.parse(expr, mode='eval').body
        match expr:
            case ast.Name():
                return PropLayout([defs[expr.id]], 'h', [None])

            case ast.BinOp(op=ast.MatMult()):
                assert isinstance(expr.right, ast.Constant)
                return PropLayout([_map_expression(expr.left, defs)], 'h', [expr.right.value])

            case _:
                return _map_expression(expr, defs)
    except SyntaxError as e:
        raise SyntaxError('Invalid layout expression') from e

PROTO_AREA_TYPES = {
    'THTPads': THTProtoArea,
    'SMDPads': SMDProtoAreaRectangles,
    'Manhattan': ManhattanProtoArea,
    'Empty': EmptyProtoArea,
}

def eval_defs(defs):
    defs = defs.replace('\n', ';')
    defs = re.sub(r'\s', '', defs)

    out = {}
    for elem in defs.split(';'):
        if not elem:
            continue

        if not (m := re.match('([a-zA-Z_][a-zA-Z0-9_]*)=([a-zA-Z_][a-zA-Z0-9_]*)\((.*)\)', elem)):
            raise SyntaxError(f'Invalid pattern definition "{elem}"')

        key, pattern, params = m.groups()
        args, kws = [], {}
        for elem in params.split(','):
            if not elem:
                continue
            if (m := re.match('([a-zA-Z_][a-zA-Z0-9_]*)=(.*)', elem)):
                param_name, param_value = m.groups()
                kws[param_name] = ast.literal_eval(param_value)

            else:
                args.append(ast.literal_eval(elem))

        out[key] = PROTO_AREA_TYPES[pattern](*args, **kws)
    return out

COMMON_DEFS = '''
empty = Empty(copper=False);
ground = Empty(copper=True);

tht = THTPads();
manhattan = Manhattan();
tht50 = THTPads(pad_size=1.0, drill=0.6, pitch=1.27);

smd100 = SMDPads(1.27, 2.54);
smd100r = SMDPads(2.54, 1.27);
smd950 = SMDPads(0.95, 2.5);
smd950r = SMDPads(2.5, 0.95);
smd800 = SMDPads(0.80, 2.0);
smd800r = SMDPads(2.0, 0.80);
smd650 = SMDPads(0.65, 2.0);
smd650r = SMDPads(2.0, 0.65);
smd500 = SMDPads(0.5, 2.0);
smd500r = SMDPads(2.0, 0.5);
'''


if __name__ == '__main__':
#    import sys
#    print('===== Layout expressions =====')
#    for line in [
#            'tht',
#            'tht@1mm',
#            'tht|tht',
#            'tht@1mm|tht',
#            'tht|tht|tht',
#            'tht@1mm|tht@2mm|tht@3mm',
#            '(tht@1mm|tht@2mm)|tht@3mm',
#            'tht@1mm|(tht@2mm|tht@3mm)',
#            'tht@2|tht|tht',
#            '(tht@1mm|tht|tht@3mm) / tht',
#            ]:
#        layout = parse_layout(line)
#        print(line, '->', layout)
#        print('    ', layout.layout(100))
#        print()
#    print('===== Pattern definitions =====')
#    for line in [
#            'tht = THTCircles()',
#            'tht = THTCircles(10)',
#            'tht = THTCircles(10, 20)',
#            'tht = THTCircles(plated=False)',
#            'tht = THTCircles(10, plated=False)',
#            ]:
#        print(line, '->', eval_defs(line))
#    print()
#    print('===== Proto board =====')
    #b = ProtoBoard('tht = THTCircles(); tht_small = THTCircles(pad_size=1.0, drill=0.6, pitch=1.27)',
    #        'tht@1in|(tht_small@2/tht@1)', mounting_holes=(3.2, 5.0, 5.0), border=2, center=False)
    #b = ProtoBoard('tht = THTCircles(); smd1 = SMDPads(2.0, 2.0); smd2 = SMDPads(0.95, 1.895); plane=Empty(copper=True)', 'tht@25mm | (smd1 + plane)', mounting_holes=(3.2, 5.0, 5.0), border=2, tight_layout=True)
    #b = ProtoBoard(COMMON_DEFS, f'((smd100 + smd100) | (smd950 + smd950) | tht50@20mm)@20mm / tht', mounting_holes=(3.2,5,5), border=1, tight_layout=True, center=True)
    b = ProtoBoard(COMMON_DEFS, f'manhattan', mounting_holes=(3.2,5,5), border=1, tight_layout=True, center=True)
    print(b.generate(80, 60))



#!/usr/bin/env python3

import os
import sys
from setuptools import setup
from setuptools.command.install import install
import subprocess
from multiprocessing import cpu_count
from pathlib import Path
import re

def get_tag():
    res = subprocess.run(['git', 'describe', '--tags', '--match', 'v*'], capture_output=True, check=True, text=True)
    return res.stdout.strip()

def get_version():
    version, _, _rest = get_tag()[1:].partition('-')
    return version

def format_readme_for_pypi():
    tag = get_tag()
    # Replace repo-relative image URLs with gitlab raw URLs. Gitlab and github render repo-relative URLs just fine, but
    # PyPI doesn't.
    return '\n'.join(
            re.sub('^.. (figure|image):: (pics/.*)$', f'.. \\1:: https://gitlab.com/gerbolyze/gerbolyze/-/raw/{tag}/\\2', line.strip('\n'))
            for line in Path('README.rst').read_text().splitlines())

setup(
    name = 'gerbolyze',
    version = get_version(),
    packages=['gerbolyze'],
    scripts=['bin/gerbolyze'],
    description = ('A high-resolution image-to-PCB converter. Gerbolyze plots SVG, PNG and JPG onto existing gerber '
        'files. It handles almost the full SVG spec and deals with text, path outlines, patterns, arbitrary paths with '
        'self-intersections and holes, etc. fully automatically. It can vectorize raster images both by contour '
        'tracing and by grayscale dithering. All processing is done at the vector level without intermediate '
        'conversions to raster images accurately preserving the input.'),
    long_description=format_readme_for_pypi(),
    long_description_content_type='text/x-rst',
    url='https://github.com/jaseg/gerbolyze',
    project_urls={
        'Source Code': 'https://git.jaseg.de/gerbolyze',
        'Bug Tracker': 'https://github.com/jaseg/gerbolyze/issues',
    },
    author = 'jaseg',
    author_email = 'gerbonara@jaseg.de',
    install_requires = ['gerbonara', 'numpy', 'python-slugify', 'lxml', 'click', 'resvg-wasi >= 0.23.0', 'svg-flatten-wasi[resvg-wasi]'],
    license = 'AGPLv3',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Manufacturing',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Religion',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'Topic :: Utilities'
        ]
)

